<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Config\MigrationActionConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationMappingConfig;
use Dedi\SyliusMigrationPlugin\Dto\MigrationOption;
use Dedi\SyliusMigrationPlugin\Registry\HandlerRegistry;
use Dedi\SyliusMigrationPlugin\Service\Logger\MigrationLogger;
use Dedi\SyliusMigrationPlugin\Service\MigrationExecutorInterface;
use Exception;
use Psr\Log\LogLevel;

class MigrationExecutor implements MigrationExecutorInterface
{
    private readonly MigrationConfig $config;

    private readonly HandlerRegistry $handlerRegistry;

    private readonly MigrationActionConfig $migrationActionConfig;

    private readonly MigrationMappingConfig $mappingRegistry;

    private readonly ReaderInterface $reader;

    private readonly WriterInterface $writer;

    private readonly MigrationLogger $migrationLogger;

    public function __construct(
        MigrationConfig $config,
        MigrationActionConfig $migrationActionConfig,
        MigrationMappingConfig $mappingRegistry,
        ReaderInterface $reader,
        WriterInterface $writer,
        MigrationLogger $migrationLogger,
        HandlerRegistry $handlerRegistry,
    ) {
        $this->config = $config;
        $this->migrationActionConfig = $migrationActionConfig;
        $this->mappingRegistry = $mappingRegistry;
        $this->reader = $reader;
        $this->writer = $writer;
        $this->migrationLogger = $migrationLogger;
        $this->handlerRegistry = $handlerRegistry;
    }

    public function execute(MigrationOption $migrationOption): void
    {
        $commit = !$migrationOption->isDryRun() && $this->config->isCommit();
        if (!$commit) {
            $this->migrationLogger->log(LogLevel::INFO, 'Running in dry mode');
            $this->defuse();
        }

        $this->migrationActionConfig->executePreMigrationActions($migrationOption);

        $numberOfTables = count($this->mappingRegistry->getTableDefinitions());
        foreach ($this->mappingRegistry->getTableDefinitions() as $index => $tableDefinition) {
            $this->migrationLogger->info(
                $tableDefinition,
                sprintf(
                    'Import from %s to %s [%s/%s]',
                    '' === $tableDefinition->getSourceTableName() ? 'SQL Query' : $tableDefinition->getSourceTableName(),
                    $tableDefinition->getTargetTableName(),
                    $index + 1,
                    $numberOfTables,
                ),
            );

            if (!$this->writer->tableExists($tableDefinition, $tableDefinition->getTargetTableName())) {
                continue;
            }

            if (!$this->reader->tableExists($tableDefinition, $tableDefinition->getSourceTableName())) {
                continue;
            }

            $offset = $tableDefinition->getLimitMappingDefinition()->getOffset();
            $step = $tableDefinition->getLimitMappingDefinition()->getStep();
            $limit = $this->config->getDebugLimitStatements() > 0 ? $this->config->getDebugLimitStatements() : $tableDefinition->getLimitMappingDefinition()->getLimit();
            $countAll = $this->reader->getResultCount($tableDefinition, $limit);
            $count = 0;

            do {
                $this->migrationLogger->info($tableDefinition, sprintf('%s/%s', $count, $countAll));

                $arrData = $this->reader->read($tableDefinition, $offset, $step);

                if (null !== $tableDefinition->getHandler()) {
                    $handler = $this->handlerRegistry->getHandler($tableDefinition->getHandler());
                    $errors = $handler->execute($arrData);

                    foreach ($errors as $error) {
                        $this->migrationLogger->error($tableDefinition, $error);
                    }
                    $count += count($arrData);
                } else {
                    foreach ($arrData as $data) {
                        try {
                            $this->writer->write($tableDefinition, $data);
                        } catch (Exception $exception) {
                            $this->migrationLogger->handleException($exception, $tableDefinition, $data);
                        }
                        ++$count;

                        if ($count === $limit) {
                            break;
                        }
                    }
                }

                $offset += $step;
            } while (0 !== count($arrData) && $count !== $limit);

            $this->writer->afterMigrate($tableDefinition);
        }

        $this->migrationActionConfig->executePostMigrationActions($commit);
    }

    private function defuse()
    {
        $this->reader->defuse();
        $this->writer->defuse();
    }
}
